import os

from pymilvus.stage.stage_operation import StageOperation

# You need to upload the local file path or folder path for import & migration & spark.
LOCAL_DIR_OR_FILE_PATH = "/Users/zilliz/Desktop/1.parquet"

# The value of the URL is fixed.
# For overseas regions, it is: https://api.cloud.zilliz.com
# For regions in China, it is: https://api.cloud.zilliz.com.cn
CLOUD_ENDPOINT = "https://api.cloud.zilliz.com"
API_KEY = "_api_key_for_cluster_org_"

# This is currently a private preview feature. If you need to use it, please submit a request and contact us.
# Before using this feature, you need to create a stage using the stage API.
STAGE_NAME = "_stage_name_for_project_"
PATH = "_path_for_stage"

def main():
    stage_operation = StageOperation(
        cloud_endpoint=CLOUD_ENDPOINT,
        api_key=API_KEY,
        stage_name=STAGE_NAME,
        path=PATH
    )
    result = stage_operation.upload_file_to_stage(LOCAL_DIR_OR_FILE_PATH)
    print(f"\nuploadFileToStage results: {result}")

if __name__ == "__main__":
    main()


import tempfile
from pathlib import Path

from pymilvus.stage.file_utils import FileUtils


def create_temp_file(path: Path, size_in_bytes: int = 10):
    path.write_bytes(b'x' * size_in_bytes)
    return path


def test_process_local_file():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_file = Path(tmp_dir) / "test.txt"
        create_temp_file(tmp_file, size_in_bytes=100)

        files, total_size = FileUtils.process_local_path(str(tmp_file))
        assert len(files) == 1
        assert Path(files[0]).resolve() == tmp_file.resolve()
        assert total_size == 100


def test_process_local_directory():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        file1 = create_temp_file(tmp_path / "file1.txt", 50)
        file2 = create_temp_file(tmp_path / "file2.txt", 70)

        subdir = tmp_path / "subdir"
        subdir.mkdir()
        file3 = create_temp_file(subdir / "file3.txt", 80)

        files, total_size = FileUtils.process_local_path(str(tmp_path))
        assert len(files) == 3
        assert total_size == 200
        resolved_files = [Path(f).resolve() for f in files]
        assert all(f.exists() for f in resolved_files)


def test_process_invalid_path():
    try:
        FileUtils.process_local_path("/path/does/not/exist")
    except ValueError as e:
        assert "Path does not exist" in str(e)
    else:
        assert False, "Expected ValueError"



import json
import logging
from typing import Optional, Union

import requests

from pymilvus.exceptions import MilvusException

logger = logging.getLogger(__name__)


def _http_headers(api_key: str):
    return {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_0) AppleWebKit/535.11 (KHTML, like Gecko) "
        "Chrome/17.0.963.56 Safari/535.11",
        "Accept": "application/json",
        "Accept-Encodin": "gzip,deflate,sdch",
        "Accept-Languag": "en-US,en;q=0.5",
        "Authorization": f"Bearer {api_key}",
    }


def _throw(msg: str):
    logger.error(msg)
    raise MilvusException(message=msg)


def _handle_response(url: str, res: json):
    inner_code = res["code"]
    if inner_code != 0:
        inner_message = res["message"]
        _throw(f"Failed to request url: {url}, code: {inner_code}, message: {inner_message}")


def _post_request(
    url: str,
    api_key: str,
    params: {},
    timeout: int = 20,
    verify: Optional[Union[bool, str]] = True,
    cert: Optional[Union[str, tuple]] = None,
    **kwargs,
) -> requests.Response:
    """Send a POST request with 1-way / 2-way optional certificate validation

    Args:
        url (str): The endpoint URL
        api_key (str): API key for authentication
        params (dict): JSON parameters for the request
        timeout (int): Timeout for the request
        verify (bool, str, optional): Either a boolean, to verify the server's TLS certificate
             or a string, which must be server's certificate path. Defaults to `True`.
        cert (str, tuple, optional): if String, path to ssl client cert file.
                                     if Tuple, ('cert', 'key') pair.

    Returns:
        requests.Response: Response object.
    """
    try:
        resp = requests.post(
            url=url,
            headers=_http_headers(api_key),
            json=params,
            timeout=timeout,
            verify=verify,
            cert=cert,
            **kwargs,
        )
        if resp.status_code != 200:
            _throw(f"Failed to post url: {url}, status code: {resp.status_code}")
        else:
            return resp
    except Exception as err:
        _throw(f"Failed to post url: {url}, error: {err}")


## bulkinsert RESTful api wrapper
def apply_stage(
    url: str,
    stage_name: str,
    path: str,
    api_key: str = "",
    verify: Optional[Union[bool, str]] = True,
    cert: Optional[Union[str, tuple]] = None,
    **kwargs,
) -> requests.Response:
    """call bulkinsert restful interface to import files

    Args:
        url (str): url of the server
        stage_name (str): name of the stage
        path (str): the path of the stage
        api_key (str): API key to authenticate your requests.
        verify (bool, str, optional): Either a boolean, to verify the server's TLS certificate
             or a string, which must be server's certificate path. Defaults to `True`.
        cert (str, tuple, optional): if String, path to ssl client cert file.
                                     if Tuple, ('cert', 'key') pair.

    Returns:
        response of the restful interface
    """
    request_url = url + "/v2/stages/apply"

    params = {
        "stageName": stage_name,
        "path": path,
    }

    options = kwargs.pop("options", {})
    if isinstance(options, dict):
        params["options"] = options

    resp = _post_request(
        url=request_url, api_key=api_key, params=params, verify=verify, cert=cert, **kwargs
    )
    _handle_response(request_url, resp.json())
    return resp



import logging
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

import urllib3
from minio import Minio
from minio.error import S3Error

from pymilvus.stage.file_utils import FileUtils
from pymilvus.stage.stage_restful import apply_stage

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class StageOperation:
    def __init__(
        self, cloud_endpoint: str, api_key: str, stage_name: str, path: Optional[str] = None
    ):
        self.cloud_endpoint = cloud_endpoint
        self.api_key = api_key
        self.stage_name = stage_name
        self.local_file_paths = []
        self.total_bytes = 0
        self.stage_info = {}
        self.path = path
        self._convert_dir_path()
        self._client = None
        self._refresh_stage_and_client()

    def _convert_dir_path(self):
        input_path = self.path
        if not input_path or input_path.endswith("/"):
            self.path = input_path
            return
        self.path = input_path + "/"

    def _refresh_stage_and_client(self):
        logger.info("refreshing stage info...")
        response = apply_stage(self.cloud_endpoint, self.stage_name, self.path, self.api_key)
        self.stage_info = response.json()["data"]
        logger.info("stage info refreshed.")

        creds = self.stage_info["credentials"]
        http_client = urllib3.PoolManager(maxsize=100)
        self._client = Minio(
            endpoint=self.stage_info["endpoint"],
            access_key=creds["tmpAK"],
            secret_key=creds["tmpSK"],
            session_token=creds["sessionToken"],
            region=self.stage_info["region"],
            secure=True,
            http_client=http_client,
        )
        logger.info("storage client refreshed")

    def _validate_size(self):
        file_size_total = self.total_bytes
        file_size_limit = self.stage_info["condition"]["maxContentLength"]
        if file_size_total > file_size_limit:
            error_message = (
                f"localFileTotalSize {file_size_total} exceeds "
                f"the maximum contentLength limit {file_size_limit} defined in the condition."
                f"If you want to upload larger files, please contact us to lift the restriction."
            )
            raise ValueError(error_message)

    def upload_file_to_stage(self, local_path: str, concurrency: int = 20):
        self.local_file_paths, self.total_bytes = FileUtils.process_local_path(local_path)
        self._validate_size()

        file_count = len(self.local_file_paths)
        logger.info(
            f"begin to upload file to stage, localDirOrFilePath:{local_path}, fileCount:{file_count} to stageName:{self.stage_name}, stagePath:{self.path}"
        )
        start_time = time.time()

        uploaded_bytes = 0
        root_path = Path(local_path).resolve()
        uploaded_bytes_lock = threading.Lock()

        def _upload_task(file_path: str, idx: int, root_path: Path):
            nonlocal uploaded_bytes
            path_obj = Path(file_path).resolve()
            if root_path.is_file():
                relative_path = path_obj.name
            else:
                relative_path = path_obj.relative_to(root_path).as_posix()

            object_name = f"{self.stage_info['uploadPath']}{relative_path}"
            file_start_time = time.time()
            try:
                size = Path(file_path).stat().st_size
                logger.info(f"uploading file, fileName:{file_path}, size:{size} bytes")
                self._put_object(file_path, object_name)
                with uploaded_bytes_lock:
                    uploaded_bytes += size
                percent = uploaded_bytes / self.total_bytes * 100
                elapsed = time.time() - file_start_time
                logger.info(
                    f"Uploaded file, {idx + 1}/{file_count}: {file_path}, elapsed:{elapsed} s, {uploaded_bytes}/{self.total_bytes} bytes, progress: {percent:.2f}%"
                )
            except S3Error as e:
                logger.error(f"Failed to upload {file_path}: {e!s}")
                raise

        with ThreadPoolExecutor(max_workers=concurrency) as executor:
            futures = []
            for i, file_path in enumerate(self.local_file_paths):
                futures.append(executor.submit(_upload_task, file_path, i, root_path))
            for f in futures:
                f.result()  # wait for all

        total_elapsed = time.time() - start_time
        logger.info(
            f"All files in {local_path} uploaded to stage, "
            f"stageName:{self.stage_info['stageName']}, stagePath: {self.path}, "
            f"totalFileCount:{file_count}, totalFileSize:{self.total_bytes}, cost time:{total_elapsed}s"
        )
        return {"stageName": self.stage_info["stageName"], "path": self.path}

    def _put_object(self, file_path: str, remote_file_path: str):
        expire_time_str = self.stage_info["credentials"]["expireTime"]
        expire_time = datetime.fromisoformat(expire_time_str.replace("Z", "+00:00"))

        now = datetime.now(timezone.utc)
        if now > expire_time:
            self._refresh_stage_and_client()

        self._upload_with_retry(file_path, remote_file_path)

    def _upload_with_retry(self, file_path: str, object_name: str, max_retries: int = 3):
        attempt = 0
        while attempt < max_retries:
            try:
                self._client.fput_object(
                    bucket_name=self.stage_info["bucketName"],
                    object_name=object_name,
                    file_path=file_path,
                )
                break
            except Exception as e:
                attempt += 1
                logger.warning(f"Attempt {attempt} failed to upload {file_path}: {e}")
                self._refresh_stage_and_client()

                if attempt == max_retries:
                    error_message = f"Upload failed after {max_retries} attempts"
                    raise RuntimeError(error_message) from e

                time.sleep(5)
